class_name Grovekeeper
extends Actor


@export var move_speed := 300.0
@export var jump_force := -400.0
@export var build_controller: BuildController

var inventory: InventoryDataResource = preload("res://scripts/inventory/player_inventory_data.tres")
var _ammo := 10 :
	set(value):
		_ammo = clampi(value, 0, 20)
		_ammo_bar.value = _ammo

@onready var rig: CanvasGroup = $Rig
@onready var animation_player: AnimationPlayer = $AnimationPlayer
@onready var projectile_shooter: ProjectileShooter = $ProjectileShooter
@onready var _health_bar: ProgressBar = $HealthBar
@onready var _health_label: Label = $HealthBar/HealthLabel
@onready var _ammo_bar: ProgressBar = $AmmoBar

# Get the gravity from the project settings to be synced with RigidBody nodes.
var gravity: int = ProjectSettings.get_setting("physics/2d/default_gravity")


func _ready() -> void:
	super._ready()
	
	_health_bar.max_value = stats.max_hp
	_health_bar.value = stats.hp
	stats.hp_changed.connect(
		func() -> void:
			_health_bar.value = stats.hp
	)
	stats.hp_changed.connect(
		func() -> void:
			_health_label.text = "HP: %d" % stats.hp
	)
	inventory.equipment_changed.connect(
		func(_slot_index: int) -> void:
			recalculate_stats()
	)
	
	projectile_shooter.actor_stats = stats


# HACK: Should be _unhandled_input
func _unhandled_input(event: InputEvent) -> void:
	if event.is_action_pressed("interact_item"):
		get_viewport().set_input_as_handled()
		var data := inventory.interact_chosen_hotbar({
			"user" : self,
			"build_controller" : build_controller,
		})
		stats.heal(data.get("hp_replenished", 0))
		_ammo += data.get("ammo_replenished", 0)
		if data.get("use_success", false):
			return
	
	if event.is_action_pressed("fire"):
		if _ammo > 0:
			get_viewport().set_input_as_handled()
			projectile_shooter.fire()
			_ammo -= 1
		return
	
	if event.is_action_pressed("hotbar_left"):
		get_viewport().set_input_as_handled()
		inventory.current_hotbar_index -= 1
	
	if event.is_action_pressed("hotbar_right"):
		get_viewport().set_input_as_handled()
		inventory.current_hotbar_index += 1


func _process(delta: float) -> void:
	# for ease of reference
	var axis = get_move_vector()
	
	# flip rig towards direction
	if axis.x != 0:
		rig.scale.x = signf(axis.x)
	
	# set animations
	if velocity.is_zero_approx():
		animation_player.current_animation = "Idle"
	else:
		animation_player.current_animation = "Walk"


func _physics_process(delta: float) -> void:
	var move_dir := get_move_vector()
	_move(move_dir.x, move_dir.y < 0, delta)
	
	projectile_shooter.look_at(get_global_mouse_position())


# ONLY TO BE CALLED IN _physics_process!
func _move(move_dir: float, jump: bool, delta: float) -> void:
	if not is_on_floor():
		velocity.y += gravity * delta
	
	if jump and is_on_floor():
		velocity.y = jump_force
	
	if not is_zero_approx(move_dir):
		if absf(move_dir) > 1:
			move_dir = signf(move_dir)
		velocity.x = move_dir * move_speed
	else:
		velocity.x = move_toward(velocity.x, 0, move_speed)
	
	move_and_slide()


func get_move_vector() -> Vector2:
	return Input.get_vector("move_left", "move_right", "move_up", "move_down")


func get_vector(axis: Vector2) -> Vector2:
	var x_vec := Grid.x_vec
	var y_vec := Grid.y_vec
	
	var vector := Vector2.ZERO
	
	vector.x = axis.x * x_vec.x + axis.y * y_vec.x
	vector.y = axis.y * y_vec.y + axis.x * x_vec.y
	
	return vector


func _on_died() -> void:
	get_tree().change_scene_to_file.call_deferred("res://rooms/main_menu.tscn")


func _on_pickupable_area_area_entered(area: Pickupable) -> void:
	area.collect(inventory)


func set_active(active):
	set_physics_process(active)
	set_process(active)
	set_process_input(active)


func recalculate_stats():
	stats.recalculate_equipment_stats(inventory.equipment_items)
