class_name InventoryDataResource
extends Resource


signal chosen_hotbar_changed(hotbar_index: int)
signal hotbar_changed(slot_index: int)
signal slot_changed(slot_index: int)
signal update_all()
signal equipment_changed(slot_index: int)

const ITEM_CAPACITY := 20
const HOTBAR_CAPACITY := 10
const EQUIPMENT_CAPACITY := 3

@export var hotbar_items: Array[ItemStackResource] = []
@export var items: Array[ItemStackResource] = []
@export var equipment_items: Array[ItemStackResource] = []

var current_hotbar_index := 0 :
	set(value):
		current_hotbar_index = wrapi(value, 0, HOTBAR_CAPACITY)
		chosen_hotbar_changed.emit(current_hotbar_index)


func _init() -> void:
	while hotbar_items.size() < HOTBAR_CAPACITY:
		hotbar_items.append(null)
	
	while items.size() < ITEM_CAPACITY:
		items.append(null)
		
	while equipment_items.size() < EQUIPMENT_CAPACITY:
		equipment_items.append(null)


func count_items(item_ids: Array) -> Dictionary:
	if not item_ids or item_ids.size() == 0:
		return {}
	
	var item_counts: Dictionary = {}
	for id in item_ids:
		item_counts[id] = 0
	
	var check := func(item_stack: ItemStackResource) -> void:
		if not is_instance_valid(item_stack) or not is_instance_valid(item_stack.item_res):
			return
		var stack_id = item_stack.item_res.id
		if item_counts.has(stack_id):
			item_counts[stack_id] += item_stack.quantity
	
	hotbar_items.map(check)
	items.map(check)
	return item_counts


func delete_items(item_count_dict: Dictionary):
	if not item_count_dict or item_count_dict.size() == 0:
		return []
	
	var delete = func(item_stack: ItemStackResource):
		if not item_stack: return
		var stack_id = item_stack.item_res.id
		if item_count_dict.has(stack_id):
			var amount = min(item_stack.quantity, item_count_dict[stack_id])
			item_stack.quantity -= amount
			item_count_dict[stack_id] -= amount
			if item_stack.quantity:
				pass
	
	hotbar_items.map(delete)
	items.map(delete)
	update_all_slots()


func update_all_slots():
	for idx in range(len(hotbar_items)):
		if hotbar_items[idx] and hotbar_items[idx].quantity <= 0:
			remove_hotbar(idx)
	for idx in range(len(items)):
		if items[idx] and items[idx].quantity <= 0:
			remove_item(idx)
	update_all.emit()


func is_hotbar_empty(slot_index: int) -> bool:
	return hotbar_items[slot_index] == null


func is_slot_empty(slot_index: int) -> bool:
	return items[slot_index] == null


func is_equipment_empty(slot_index: int) -> bool:
	return equipment_items[slot_index] == null


func get_chosen_hotbar() -> ItemStackResource:
	return get_hotbar(current_hotbar_index)


func _contain_same_item(lhs_stack: ItemStackResource, rhs_stack: ItemStackResource) -> bool:
	if lhs_stack == null or rhs_stack == null:
		return lhs_stack == null and rhs_stack == null
	if lhs_stack.item_res == null or rhs_stack.item_res == null:
		return lhs_stack.item_res == null and rhs_stack.item_res == null
	return lhs_stack.item_res.id == rhs_stack.item_res.id


func append_item_stack_to_available_stack(item_stack: ItemStackResource) -> bool:
	for index in range(ITEM_CAPACITY):
		if _contain_same_item(items[index], item_stack):
			var moved_item: int = min(item_stack.quantity, items[index].item_res.max_stack - items[index].quantity)
			
			items[index].quantity += moved_item
			item_stack.quantity -= moved_item
			
			if moved_item > 0:
				slot_changed.emit(index)
			if item_stack.quantity == 0:
				return true
	
	for index in range(HOTBAR_CAPACITY):
		if _contain_same_item(hotbar_items[index], item_stack):
			var moved_item: int = min(item_stack.quantity, hotbar_items[index].item_res.max_stack - hotbar_items[index].quantity)
			
			hotbar_items[index].quantity += moved_item
			item_stack.quantity -= moved_item
			
			if moved_item > 0:
				hotbar_changed.emit(index)
			if item_stack.quantity == 0:
				return true
	return false


func append_item_stack_to_inventory(item_stack: ItemStackResource) -> bool:
	if append_item_stack_to_available_stack(item_stack):
		return true
	if append_item_to_empty_slot(item_stack):
		return true
	# TODO: handle if full (ex. add message) 
	return false


func insert_item_stack_to_inventory(item_stack: ItemStackResource) -> bool:
	# TODO: handle if full (ex. add message) 
	return append_item_stack_to_inventory(item_stack)


# Add item resource to inventory
# Returns:
#   True if the item is successfully added and False otherwise.
func insert_item_to_inventory(item: ItemResource, quantity: int = 1) -> bool:
	var item_stack = ItemStackResource.new()
	item_stack.init_stack(item, quantity)

	# TODO: handle if full (ex. add message) 
	return append_item_stack_to_inventory(item_stack)


# TBD: possible method when using an item, a signal is emitted from the item stack to inventory and then to ui
func interact_chosen_hotbar(data: Dictionary) -> Dictionary:
	var ret_data := {}
	var item_stack: ItemStackResource = get_chosen_hotbar()
	var interacted_hotbar_index: int = current_hotbar_index
	if item_stack:
		if item_stack.quantity > 0:
			ret_data = item_stack.interact(data)
			hotbar_changed.emit(interacted_hotbar_index)
		if item_stack.quantity <= 0:
			remove_hotbar(interacted_hotbar_index)
	
	return ret_data


func get_hotbar(slot_index: int) -> ItemStackResource:
	return hotbar_items[slot_index]


func get_item(slot_index: int) -> ItemStackResource:
	return items[slot_index]


func get_equipment(slot_index: int) -> ItemStackResource:
	return equipment_items[slot_index]


func append_item_to_empty_slot(item: ItemStackResource) -> bool:
	for index in range(HOTBAR_CAPACITY):
		if is_hotbar_empty(index):
			insert_hotbar(item, index)
			return true
	
	for index in range(ITEM_CAPACITY):
		if is_slot_empty(index):
			insert_item(item, index)
			return true
	
	return false


func insert_hotbar(item: ItemStackResource, slot_index: int) -> void:
	hotbar_items[slot_index] = item
	hotbar_changed.emit(slot_index)


func insert_item(item: ItemStackResource, slot_index: int) -> void:
	items[slot_index] = item
	slot_changed.emit(slot_index)


func insert_equipment(item: ItemStackResource, slot_index: int) -> void:
	equipment_items[slot_index] = item
	equipment_changed.emit(slot_index)


func remove_hotbar(slot_index: int) -> void:
	hotbar_items[slot_index] = null
	hotbar_changed.emit(slot_index)


func remove_item(slot_index: int) -> void:
	items[slot_index] = null
	slot_changed.emit(slot_index)


func remove_equipment(slot_index: int) -> void:
	equipment_items[slot_index] = null
	equipment_changed.emit(slot_index)

